/***********************************************************************************
Class SC88_Display

  This class object is responsible for painting an approximate representation of
a Roland Sound Canvas 88 LCD (plasma display). It also does some MIDI System Exclusive
(SYSEX) generation for sending MIDI messages to an actual SC88.

  The font functions are from Charles Petzold's book "Programming Windows95".

      You may freely distribute, modify, and use this source
    code for non-commercial purposes, so long as this notice
    is left intact.  Commercial use of this source code or 
    any part of it requires express written permission from
    Michael Leach <mike@pcez.com>

 **********************************************************************************/

#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include "SC88.hpp"
#include "ezfont.h"

#define  LCDFont        EzCreateFont (hdc, "LCD Display Grid", 130, 0, 0, TRUE)
#define  SmVerdanaFont  EzCreateFont (hdc, "Verdana", 60, 0, 0, TRUE)
#define  MedVerdanaFont EzCreateFont (hdc, "Verdana", 140, 0, 0, TRUE)
#define  LrgVerdanaFont EzCreateFont (hdc, "Verdana", 180, 0, 0, TRUE)

unsigned char SYSEX_HDR[8]  = {0xF0, 0x41, 0x10, 0x45, 0x12, 0x10, 1, 0};
unsigned char STRING_HDR[8] = {0xF0, 0x41, 0x10, 0x45, 0x12, 0x10, 0, 0};
FILE *f;

SC88_Display::SC88_Display(){
   device      = 0;

   gs.part     = 1;
   gs.instr    = 1;
   strcpy(gs.patchName, "Piano 01");
   gs.level    = 100;
   gs.pan      = 0;
   gs.reverb   = 40;
   gs.chorus   = 0;
   gs.keyShift = 0;
   gs.MIDIBank = 'A';
   gs.MIDIChan = 1;

   rect.left   = SC_LEFT;
   rect.top    = SC_TOP;
   rect.right  = SC_LEFT + DISPLAY_X;
   rect.bottom = SC_TOP  + DISPLAY_Y;

   grid.left   = SC_LEFT + 90;
   grid.top    = SC_TOP  + 23;
   grid.right  = SC_LEFT + DISPLAY_X - 4;
   grid.bottom = SC_TOP  + DISPLAY_Y - 5;

   bkBrush      = CreateSolidBrush(RGB(255,255,0)) ;
   gridOffBrush = CreateSolidBrush(RGB(192,192,192)) ;
   gridOnBrush  = CreateSolidBrush(RGB(0,  0,  0)) ;

   for(int i = 0; i < 256; i++)
      dotOn[i] = 0;
}

SC88_Display::~SC88_Display(){
   midiOutClose(hmo);
}

void SC88_Display::OpenMIDIOut(void){
   midiOutOpen(&hmo, device, 0L, 0L, 0L);
}

void SC88_Display::CloseMIDIOut(void){
   midiOutClose(hmo);
}

void SC88_Display::DrawDisplay(void){
   HDC         hdc ;
   HPEN        hPen;
   PAINTSTRUCT ps ;
   RECT        *tmp;
   char        str[80];
   int         y      = 5;
   int         row    = 0,
               column = 0,
               i;

   hdc = BeginPaint (hwnd, &ps) ;

   SetBkColor(hdc, RGB(255, 255, 0));

   FillRect (hdc, &rect, bkBrush);

// Display grid lines

   for(i = 0; i < 256; i++){
      tmp = &dot[i];
      tmp->left   = grid.left + (column * 16);
      tmp->right  = grid.left + ((column * 16) + 15);
      tmp->top    = grid.top  + (row * 7);
      tmp->bottom = grid.top  + ((row * 7) + 6);

      if(dotOn[i])
         FillRect (hdc, tmp, gridOnBrush);
      else
         FillRect (hdc, tmp, gridOffBrush);

      if(++column >= 16){
         column = 0;
         row++;
      }
   }

// White border light effect
#define PAD    2
   hPen = CreatePen(PS_SOLID, 4, RGB(0, 0, 0));
   SelectObject(hdc, hPen);
   MoveToEx(hdc, rect.left - PAD, rect.bottom + PAD, NULL);
   LineTo(hdc, rect.left  - PAD,  rect.top - PAD);
   LineTo(hdc, rect.right + PAD,  rect.top - PAD);

   hPen = CreatePen(PS_SOLID, 1, RGB(193, 193, 193));
   SelectObject(hdc, hPen);

   for(i = 0; i < 4; i++){
      MoveToEx(hdc, rect.left - i,  rect.bottom + i, NULL);
      LineTo(hdc,  rect.right + i,  rect.bottom + i);
      LineTo(hdc,  rect.right + i,  rect.top - i);
   }

// LCD Font Output
   SelectObject (hdc, LCDFont) ;
   wsprintf(str, "%c%02u %03u %s", gs.MIDIBank, gs.MIDIChan, gs.instr,
      gs.patchName);
   TextOut(hdc, rect.left + 5, rect.top + y, str, strlen(str));
   y += Y_INC;
   SetTextAlign(hdc, TA_TOP | TA_RIGHT);

   wsprintf(str, "%u", gs.level);
   TextOut(hdc, rect.left + 38, rect.top + y, str, strlen(str));

   wsprintf(str, "%u", gs.pan);
   TextOut(hdc, rect.left + 82, rect.top + y, str, strlen(str));
   y += Y_INC;

   wsprintf(str, "%u", gs.reverb);
   TextOut(hdc, rect.left + 38, rect.top + y, str, strlen(str));

   wsprintf(str, "%u", gs.chorus);
   TextOut(hdc, rect.left + 82, rect.top + y, str, strlen(str));
   y += Y_INC;

   wsprintf(str, "%u", gs.keyShift);
   TextOut(hdc, rect.left + 38, rect.top + y, str, strlen(str));

   wsprintf(str, "%c%02u", gs.MIDIBank, gs.MIDIChan);
   TextOut(hdc, rect.left + 82, rect.top + y, str, strlen(str));

// Sound Pen and Digital Passsage Logo Text
   SelectObject (hdc, LrgVerdanaFont) ;
   SetTextAlign(hdc, TA_TOP | TA_LEFT);
   SetBkMode(hdc, TRANSPARENT);
   wsprintf(str, "%s", "Sound Pen");
   TextOut(hdc, SC_LEFT+DISPLAY_X + 85, 160, str, strlen(str));
   SetTextColor(hdc, RGB(192, 192, 192));
   TextOut(hdc, SC_LEFT+DISPLAY_X + 83, 158, str, strlen(str));

   SelectObject (hdc, MedVerdanaFont);
   SetTextColor(hdc, 0);
   wsprintf(str, "%s", "Digital Passage");
   TextOut(hdc, SC_LEFT+DISPLAY_X + 85, 190, str, strlen(str));
   SetTextColor(hdc, RGB(192, 192, 192));
   TextOut(hdc, SC_LEFT+DISPLAY_X + 83, 188, str, strlen(str));
   SetTextColor(hdc, 0);

// Small Verdana Text
   y = 25;
   SelectObject (hdc, SmVerdanaFont) ;
   wsprintf(str, "%s", "PART");
   TextOut(hdc, rect.left + 5, rect.top - 15, str, strlen(str));
   wsprintf(str, "%s", "INSTRUMENT");
   TextOut(hdc, rect.left + 90, rect.top - 15, str, strlen(str));
   wsprintf(str, "%s", "MIDI OUT DEVICE");
   TextOut(hdc, SC_LEFT+DISPLAY_X + 85, 10, str, strlen(str));
   wsprintf(str, "%s", "DISPLAY STRING");
   TextOut(hdc, SC_LEFT, SC_TOP+DISPLAY_Y + 20, str, strlen(str));

   SetBkMode(hdc, OPAQUE);
   SetBkColor(hdc, RGB(255, 255, 0));
   wsprintf(str, "%s", "LEVEL");
   TextOut(hdc, rect.left + 5, rect.top + y, str, strlen(str));
   wsprintf(str, "%s", "PAN");
   TextOut(hdc, rect.left + 48, rect.top + y, str, strlen(str));
   y += Y_INC;

   wsprintf(str, "%s", "REVERB");
   TextOut(hdc, rect.left + 5, rect.top + y, str, strlen(str));
   wsprintf(str, "%s", "CHORUS");
   TextOut(hdc, rect.left + 48, rect.top + y, str, strlen(str));
   y += Y_INC;

   wsprintf(str, "%s", "K SHIFT");
   TextOut(hdc, rect.left + 5, rect.top + y, str, strlen(str));
   wsprintf(str, "%s", "MIDI CH");
   TextOut(hdc, rect.left + 48, rect.top + y, str, strlen(str));

   EndPaint (hwnd, &ps) ;
}

void SC88_Display::UpdateGrid(POINT pt){
   if(PtInRect(&grid, pt)){
      for(int i = 0; i < 256; i++){
         if(PtInRect(&dot[i], pt)){
            dotOn[i] = 1 - dotOn[i]; // Flip switch if mouse click is in a valid rectangle.
            InvalidateRect(hwnd, &dot[i], TRUE);
            break;
         }
      }
   }
}

void SC88_Display::AllOn(void){
   for(int i = 0; i < 256; i++)
      dotOn[i] = 1;
   InvalidateRect(hwnd, &grid, TRUE);
}

void SC88_Display::AllOff(void){
   for(int i = 0; i < 256; i++)
      dotOn[i] = 0;
   InvalidateRect(hwnd, &grid, TRUE);
}

#define DATA_START      8
void SC88_Display::GenerateDisplayMsg(int action){
   int  i, j;
   int  result;
   char errStr[80];

   memset(msg, 0, DISPLAY_MSG_SIZE);
   memcpy(msg, SYSEX_HDR, 8);

   i = 0;
   for(j = 0; j < 16; j++){
      if(dotOn[i])
         msg[DATA_START + j] |= 16;
      if(dotOn[i + 1])
         msg[DATA_START + j] |= 8;
      if(dotOn[i + 2])
         msg[DATA_START + j] |= 4;
      if(dotOn[i + 3])
         msg[DATA_START + j] |= 2;
      if(dotOn[i + 4])
         msg[DATA_START + j] |= 1;
      i += 16;
   }
   i = 5;
   for(j = 16; j < 32; j++){
      if(dotOn[i])
         msg[DATA_START + j] |= 16;
      if(dotOn[i + 1])
         msg[DATA_START + j] |= 8;
      if(dotOn[i + 2])
         msg[DATA_START + j] |= 4;
      if(dotOn[i + 3])
         msg[DATA_START + j] |= 2;
      if(dotOn[i + 4])
         msg[DATA_START + j] |= 1;
      i += 16;
   }

   i = 10;
   for(j = 32; j < 48; j++){
      if(dotOn[i])
         msg[DATA_START + j] |= 16;
      if(dotOn[i + 1])
         msg[DATA_START + j] |= 8;
      if(dotOn[i + 2])
         msg[DATA_START + j] |= 4;
      if(dotOn[i + 3])
         msg[DATA_START + j] |= 2;
      if(dotOn[i + 4])
         msg[DATA_START + j] |= 1;
      i += 16;
   }

   i = 15;
   for(j = 48; j < 64; j++){
      if(dotOn[i])
         msg[DATA_START + j] |= 16;
      i += 16;
   }

   msg[DISPLAY_MSG_SIZE - 1] = END_SYSEX;

   result = SYSEXCheckSum(msg);
   if(result < 0){
      wsprintf(errStr, "Bad Checksum Error. Result = %d", result);
      MessageBox(NULL, errStr, "Whoops!", MB_OK);
      return;
   }
   if(action == SC_SEND)
      SendMessage(msg, DISPLAY_MSG_SIZE);
   else if(action = SC_SAVE){
      f = fopen("DOTS.SYX", "wb");
      fwrite(msg, result, 1, f);
      fclose(f);
   }
}

void SC88_Display::GenerateStringMsg(char *str, int action){
   unsigned char tmp[80];
   unsigned char *ptr;
   int i , size;

   memset(tmp, 0, 80);
   memcpy(tmp, STRING_HDR, 8);
   size = strlen(str);
   ptr  = tmp;
   ptr += 8;
   for(i = 0; i < size; i++){
      *ptr = str[i];
      ptr++;
   }

   *ptr = 0;    ptr++;
   *ptr = 0xF7;
   size = SYSEXCheckSum(tmp);

   if(action == SC_SEND)
      SendMessage(tmp, size);
   else if(action == SC_SAVE){
      f = fopen("TEXT.SYX", "wb");
      fwrite(tmp, size, 1, f);
      fclose(f);
   }
}

void SC88_Display::SendMessage(unsigned char* message, UINT size){
   hdr.lpData         = (char*)message;
   hdr.dwBufferLength = size;
   hdr.dwFlags        = 0;

   midiOutPrepareHeader(hmo, (LPMIDIHDR)&hdr, sizeof(MIDIHDR));
   midiOutLongMsg(hmo, (LPMIDIHDR)&hdr, sizeof(MIDIHDR));
   midiOutUnprepareHeader(hmo, (LPMIDIHDR)&hdr, sizeof(MIDIHDR));
}

/*
 Input: *msg points to first byte of sysex message. The byte before the
 end SYSEX byte (0xF7) is assumed to be 0. This byte is then set to the
 value of the calculated checksum.

 Returns -1 if invalid SYSEX message.
 Returns -2 if the size of the message exceeds MAX_MSG (usually 256 bytes).
 Returns size of message in bytes if valid.

 Roland SC-88 Sound Canvas specific. Should be applicable to most Roland
 checksum calculations.                                    */

int SC88_Display::SYSEXCheckSum(unsigned char* message){
   int  sum   = 0,
        count = 0;
   unsigned char *ptr;

   ptr = message;
   if(*ptr != BEGIN_SYSEX)
      return -1;

   ptr += DATA_OFFSET;

   //Sum up all the data bytes.
   while(*ptr != END_SYSEX){
      sum += *ptr;
      if(++count >= MAX_MSG)
         return -2;
      ptr++;
   }
   count++;

   // Calculate the check sum, write it to the message array and return
   // the number of data bytes in the message.
   while(sum > 128) sum -= 128;
   ptr--;
   *ptr = 128 - sum;
   return(count + DATA_OFFSET);
}



